/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/

#include "SDL_direct3dtexturegrid.h"


HRESULT TextureGrid_create(_THIS,
						   int imageWidth, int imageHeight,
						   int maxCellWidth, int maxCellHeight,
						   struct TextureGrid **output)
{
	struct TextureGrid *grid;
	int x, y;

	/* Basic error checking */
	if ( this == NULL ||
		 output == NULL ||
		 maxCellWidth == 0 ||
		 maxCellHeight == 0 ) {
		return SDLD3DERR_INVALIDPARAMS;
	}

	/* Allocate the grid */
	grid = (struct TextureGrid *) malloc(sizeof(struct TextureGrid));
	if ( grid == NULL ) {
		/* SDL_OutOfMemory(); */
		return E_OUTOFMEMORY;
	}
	memset(grid, 0, sizeof(struct TextureGrid));

	/* Initialize the number of cells */
	grid->columns = imageWidth / maxCellWidth;
	if ( (imageWidth % maxCellWidth) > 0 ) {
		grid->columns++;
	}

	grid->rows = imageHeight / maxCellHeight;
	if ( (imageHeight % maxCellHeight) > 0 ) {
		grid->rows++;
	}

	/* Initialize the texture width and height */
	grid->texture_width = maxCellWidth;
	grid->texture_height = maxCellHeight;

	/* Initialize the cell columns */
	grid->cell_widths = (int *) malloc(grid->columns * sizeof(int));
	if ( grid->cell_widths == NULL ) {
		/* SDL_OutOfMemory(); */
		return E_OUTOFMEMORY;
	}

	for ( x = 0; x < grid->columns; x++ ) {
		grid->cell_widths[x] = maxCellWidth;
	}
	if ( (imageWidth % maxCellWidth) > 0 ) {
		grid->cell_widths[grid->columns - 1] = imageWidth - (maxCellWidth * (grid->columns - 1));
	}

	/* Initialize the cell rows */
	grid->cell_heights = (int *) malloc(grid->rows * sizeof(int));
	if ( grid->cell_heights == NULL ) {
		/* SDL_OutOfMemory(); */
		return E_OUTOFMEMORY;
	}

	for ( y = 0; y < grid->rows; y++ ) {
		grid->cell_heights[y] = maxCellHeight;
	}
	if ( (imageHeight % maxCellHeight) > 0 ) {
		grid->cell_heights[grid->rows - 1] = imageHeight - (maxCellHeight * (grid->rows - 1));
	}

	/* Create textures */
	grid->textures = (IDirect3DTexture9 ***) malloc(grid->columns * sizeof(IDirect3DTexture9 **));
	if ( grid->textures == NULL ) {
		/* SDL_OutOfMemory(); */
		return E_OUTOFMEMORY;
	}

	for ( x = 0; x < grid->columns; x++ ) {
		grid->textures[x] = (IDirect3DTexture9 **) malloc(grid->rows * sizeof(IDirect3DTexture9 *));
		if ( grid->textures[x] == NULL ) {
			/* SDL_OutOfMemory(); */
			return E_OUTOFMEMORY;
		}

		for ( y = 0; y < grid->rows; y++ ) {
			/* Create a texture object */
			VERIFY_FCALL(IDirect3DDevice9_CreateTexture, (
				/* Our global Direct3D device, created by DIRECT3D_SetVideoMode */
				d3d9_device,
				/* Texture dimensions */
				maxCellWidth, maxCellHeight,
				/* Mip levels (TODO, davidl: Investigate adjusting this) */
				0,
				/* Usage */
				0,
				/* Pixel format (TODO, davidl: Investigate adjusting this) */
				D3DFMT_A8R8G8B8,
				/* Pool (memory class) (TODO, davidl: Investigate adjusting this) */
				D3DPOOL_MANAGED,
				/* Texture object */
				&grid->textures[x][y],
				/* Handle, always set to NULL */
				NULL),
				/* return the offending hresult on error */
				return _fcall_hresult);

#if 0
			/* Create a texture object */
			VERIFY_FCALL(D3DXCreateTexture, (
				/* Our global, Direct3D device, created by DIRECT3D_SetVideoMode */
				d3d9_device,
				/* Texture dimensions. */
				maxCellWidth, maxCellHeight,
				/* Mip Levels */
				D3DX_DEFAULT,
				/* Usage */
				0,
				/* Pixel format */
				D3DFMT_A8R8G8B8,
				/* Pool (memory class) */
				D3DPOOL_MANAGED,
				/* Texture object */
				&grid->textures[x][y]),
				/* return the offending hresult on error */
				return _fcall_hresult);
#endif
		}
	}

	/* We're done! */
	*output = grid;
	return S_OK;
}

void TextureGrid_destroy(struct TextureGrid *grid)
{
	int column, row;

	/* Destroy Textures and their containing structures */
	for ( column = 0; column < grid->columns; column++ ) {
		for ( row = 0; row < grid->rows; row++ ) {
			IDirect3DTexture9_Release(grid->textures[column][row]);
		}

		free(grid->textures[column]);
	}
	free(grid->textures);

	/* Destroy cell size arrays */
	free(grid->cell_heights);
	free(grid->cell_widths);

	/* Destroy the grid structure */
	free(grid);
}

HRESULT TextureGrid_copyPixelsFromSurface(struct TextureGrid *grid,
										  SDL_Surface *srcData)
{
	int column, row;
	int srcXpos, srcYpos;
	int yTmp, yBottom;
	D3DLOCKED_RECT lockedRect;
	RECT destRect;
	size_t copySize;
	Uint8 *copySrc, *copyDest;

	srcXpos = 0;
	for ( column = 0; column < grid->columns; column++ ) {
		srcYpos = 0;
		for ( row = 0; row < grid->rows; row++ ) {
			destRect.left = destRect.top = 0;
			destRect.right = grid->cell_widths[column];
			destRect.bottom = grid->cell_heights[row];

			VERIFY_FCALL(IDirect3DTexture9_LockRect, (
				/* Texture to lock/copy-to */
				grid->textures[column][row],
				/* Level */
				0,
				/* Locked rect structure, which will contain the pointer to the
				pixel buffer */
				&lockedRect,
				/* Area to lock (lock everything) */
				&destRect,
				/* Flags */
				0),
				/* return failed HRESULT on error */
				return _fcall_hresult);

			//memset(lockedRect.pBits, 0xFFFFFFFF, (lockedRect.Pitch * 64));

			yBottom = grid->cell_heights[row];
			copySize = srcData->format->BytesPerPixel * grid->cell_widths[column];
			for ( yTmp = 0; yTmp < yBottom; yTmp++ ) {
				copySrc = ((Uint8 *)srcData->pixels) + (((srcYpos + yTmp) * srcData->pitch) + (srcXpos * srcData->format->BytesPerPixel));
				copyDest = ((Uint8 *)lockedRect.pBits) + (yTmp * lockedRect.Pitch);
				memcpy(copyDest, copySrc, copySize);
			}

			VERIFY_FCALL(IDirect3DTexture9_UnlockRect, (
				/* The texture to unlock */
				grid->textures[column][row],
				/* Level */
				0),
				/* return failed HRESULT on error */
				return _fcall_hresult);

			srcYpos += grid->cell_heights[row];
		}

		srcXpos += grid->cell_widths[column];
	}

	return S_OK;
}

/* Renders a single texture */
static HRESULT render_texture(IDirect3DDevice9 *device,
							  IDirect3DTexture9 *texture,
							  int x, int y,
							  int width, int height,
							  float u1, float v1, float u2, float v2,
							  Uint8 surface_alpha)
{
	/* Prevent rounding errors */
	float realX = (float)x + 0.5f;
	float realY = (float)y + 0.5f;

	/* Set up the alpha value */
	Uint32 color = surface_alpha << 24;

	/* Define a quad by using a triangle strip */
	struct textured_vertex vertices[] = {
		{ realX,			realY,			0,	1,	color,	u1,	v1 },
		{ realX + width,	realY,			0,	1,	color,	u2,	v1 },
		{ realX,			realY + height,	0,	1,	color,	u1,	v2 },
		{ realX + width,	realY + height,	0,	1,	color,	u2,	v2 }
	};

	/* Set the device's texture */
	VERIFY_FCALL(IDirect3DDevice9_SetTexture, (
		/* Device to render with */
		device,
		/* Stage */
		0,
		/* Texture */
		(IDirect3DBaseTexture9 *) texture),
		/* return generated error code on failure */
		return _fcall_hresult);

	/* Render the quad, using the just-set texture */
	VERIFY_FCALL(IDirect3DDevice9_DrawPrimitiveUP, (
		/* Device to render with */
		device,
		/* PrimitiveType */
		D3DPT_TRIANGLESTRIP,
		/* PrimitiveCount (number of triangles) */
		2,
		/* pVertexStreamZeroData (triangle data) */
		vertices,
		/* VertexStreamZeroStride */
		sizeof(struct textured_vertex)),
		/* return generated error code on failure */
		return _fcall_hresult);

	return S_OK;
}

/* Clips one rect to another */
static void clip_to_rect(SDL_Rect *rect, const SDL_Rect *clip_rect)
{
	int dx, dy;
	int rx, ry, rw, rh;

	rx = rect->x;
	ry = rect->y;
	rw = rect->w;
	rh = rect->h;

	dx = clip_rect->x - rx;
	if(dx > 0) {
		rw -= dx;
		rx += dx;
	}
	dx = rx + rw - clip_rect->x - clip_rect->w;
	if(dx > 0)
		rw -= dx;

	dy = clip_rect->y - ry;
	if(dy > 0) {
		rh -= dy;
		ry += dy;
	}
	dy = ry + rh - clip_rect->y - clip_rect->h;
	if(dy > 0)
		rh -= dy;

	rect->x = rx;
	rect->y = ry;
	rect->w = (rw > 0) ? rw : 0;
	rect->h = (rh > 0) ? rh : 0;
}

HRESULT TextureGrid_draw(struct TextureGrid *grid,
						 IDirect3DDevice9 *device,
						 SDL_Rect *src_rect,
						 SDL_Rect *dest_rect,
						 Uint8 src_surface_alpha)
{
	/* Current cell */
	int column, row;

	/* Offset into the unclipped src_rect */
	int cellOffsetX, cellOffsetY;

	/* Current destination */
	int destX, destY;

	/* Texture coordinates */
	float u1, v1, u2, v2;

	/* Used for calculating texture coordinates */
	SDL_Rect cell_rect, src_on_cell_rect;

	cellOffsetX = 0;
	destX = dest_rect->x;
	for ( column = 0; column < grid->columns; column++ ) {
		cellOffsetY = 0;
		destY = dest_rect->y;

		for ( row = 0; row < grid->rows; row++ ) {
			/* Calculate the texture coordinates */
			cell_rect.x = cellOffsetX;
			cell_rect.y = cellOffsetY;
			cell_rect.w = grid->cell_widths[column];
			cell_rect.h = grid->cell_heights[row];

			memcpy(&src_on_cell_rect, src_rect, sizeof(SDL_Rect));
			clip_to_rect(&src_on_cell_rect, &cell_rect);

			if (src_on_cell_rect.w <= 0 || src_on_cell_rect.h <= 0)
			{
				cellOffsetY += grid->cell_heights[row];

				/* Cells that do not exist in src_rect do not need to be drawn */
				continue;
			}

			u1 = (float)(src_on_cell_rect.x - cellOffsetX) / grid->texture_width;
			v1 = (float)(src_on_cell_rect.y - cellOffsetY) / grid->texture_height;
			u2 = (float)(src_on_cell_rect.x + src_on_cell_rect.w - cellOffsetX) / grid->texture_width;
			v2 = (float)(src_on_cell_rect.y + src_on_cell_rect.h - cellOffsetY) / grid->texture_height;

			/* Render the current cell's texture */
			VERIFY_FCALL(render_texture, (
				/* Device to render with */
				device,
				/* Texture */
				grid->textures[column][row],
				/* Position */
				destX, destY,
				/* Dimensions (of the drawable portion of the texture) */
				src_on_cell_rect.w, src_on_cell_rect.h,
				/* Texture coordinates */
				u1, v1, u2, v2,
				/* Per-surface alpha */
				src_surface_alpha),
				/* return generated error code on failure */
				return _fcall_hresult);

			cellOffsetY += grid->cell_heights[row];
			destY += src_on_cell_rect.h;
		}

		cellOffsetX += grid->cell_widths[column];
		destX += src_on_cell_rect.w;
	}

	/* All done! */
	return S_OK;
}

